﻿"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.ErrorsParameterParser = void 0;
const tslib_1 = require("tslib");
const common_1 = require("@kitajs/common");
const node_http_1 = tslib_1.__importDefault(require("node:http"));
const ts_json_schema_generator_1 = require("ts-json-schema-generator");
const helpers_1 = require("../schema/helpers");
const nodes_1 = require("../util/nodes");
// A map of @fastify/sensible error names to their codes
const messageStatusMap = {};
for (const key in node_http_1.default.STATUS_CODES) {
    messageStatusMap[normalize(key, node_http_1.default.STATUS_CODES[key])] = key;
}
/**
 * Normalize function used by fastify sensible.
 *
 * @link https://github.com/fastify/fastify-sensible/blob/a2f544d665d8d68c7683d94ee34fd08d9ecf3301/lib/httpErrors.js#L11C10-L11C20
 */
function normalize(code, msg) {
    if (code === '414')
        return 'uriTooLong';
    if (code === '505')
        return 'httpVersionNotSupported';
    msg = msg.split(' ').join('').replace(/'/g, '');
    msg = msg[0].toLowerCase() + msg.slice(1);
    return msg;
}
class ErrorsParameterParser {
    collector;
    constructor(collector) {
        this.collector = collector;
    }
    supports(param) {
        return (0, nodes_1.getTypeNodeName)(param) === 'HttpErrors';
    }
    parse(param, route, fn) {
        // Adds fastify sensible plugin
        if (!this.collector.getPlugin('fastifySensible')) {
            this.collector.addPlugin('fastifySensible', {
                name: 'fastifySensible',
                importUrl: '@fastify/sensible',
                options: { sharedSchemaId: 'HttpError' }
            });
        }
        // To find throw declarations
        const paramValue = (0, nodes_1.getTypeName)(param);
        if (!fn.body) {
            throw new common_1.InvalidParameterUsageError('This function has no body', fn.name || fn);
        }
        ts_json_schema_generator_1.ts.forEachChild(fn.body, function loopSourceNodes(node) {
            if (
            // Checks first if the node is a throw statement
            ts_json_schema_generator_1.ts.isThrowStatement(node) &&
                // is a method call, like `throw [errors.notFound()]`
                ts_json_schema_generator_1.ts.isCallExpression(node.expression) &&
                // is a property access expression, like `throw [errors.notFound]()`
                ts_json_schema_generator_1.ts.isPropertyAccessExpression(node.expression.expression) &&
                // is a property access expression, like `throw [errors].notFound()`
                ts_json_schema_generator_1.ts.isIdentifier(node.expression.expression.expression) &&
                // is the same as the parameter name
                node.expression.expression.expression.escapedText === paramValue) {
                // The method name, like `throw errors.[notFound]()`
                const method = node.expression.expression.name.escapedText;
                const status = messageStatusMap[method];
                if (!status) {
                    throw new common_1.UnknownHttpError(node.expression.expression.name);
                }
                // Adds the route schema
                (0, helpers_1.mergeSchema)(route, {
                    response: {
                        [status]: {
                            // This type will be registered by `sharedSchemaId` option
                            // from @fastify/sensible
                            $ref: 'HttpError'
                        }
                    }
                });
                return;
            }
            ts_json_schema_generator_1.ts.forEachChild(node, loopSourceNodes);
        });
        return {
            name: ErrorsParameterParser.name,
            value: `${common_1.kFastifyParam}.httpErrors`
        };
    }
}
exports.ErrorsParameterParser = ErrorsParameterParser;
//# sourceMappingURL=errors.js.map