﻿"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.KitaParser = void 0;
const tslib_1 = require("tslib");
const common_1 = require("@kitajs/common");
const path_1 = tslib_1.__importDefault(require("path"));
const ts_json_schema_generator_1 = require("ts-json-schema-generator");
const parameter_parsers_1 = require("./parameter-parsers");
const provider_parsers_1 = require("./provider-parsers");
const route_parsers_1 = require("./route-parsers");
const builder_1 = require("./schema/builder");
const nodes_1 = require("./util/nodes");
const paths_1 = require("./util/paths");
const traverser_1 = require("./util/traverser");
class KitaParser {
    config;
    program;
    formatter;
    providers = new Map();
    routes = new Map();
    plugins = new Map();
    schemaBuilder;
    rootRouteParser;
    rootParameterParser;
    rootProviderParser;
    /**
     * Creates a KitaParser instance with the given config.
     *
     * You may pass a formatter to format the source code of the routes and providers.
     */
    static create(config, compilerOptions, rootNames, formatter) {
        const parser = new KitaParser(config, ts_json_schema_generator_1.ts.createProgram({
            options: compilerOptions,
            // Adds both providers and controllers
            rootNames: rootNames
        }), formatter);
        return parser;
    }
    static createWatcher(config, compilerOptions, formatter) {
        const parser = {};
        let watcher;
        // Note that there is another overload for `createWatchCompilerHost` that takes
        // a set of root files.
        const host = ts_json_schema_generator_1.ts.createWatchCompilerHost(config.tsconfig, compilerOptions, ts_json_schema_generator_1.ts.sys, 
        // No emit is needed
        ts_json_schema_generator_1.ts.createSemanticDiagnosticsBuilderProgram, 
        // Do not care if typescript code has problems
        () => { }, (diagnostic) => {
            if (
            // Compilation successful
            diagnostic.code !== 6194 &&
                // Found X errors
                diagnostic.code !== 6193) {
                return;
            }
            // Watcher not initialized yet
            if (!watcher) {
                return;
            }
            const program = watcher.getProgram().getProgram();
            try {
                parser.ref = new KitaParser(config, program, parser.ref.formatter);
            }
            catch (error) {
                const promise = parser.onChange?.(parser.ref);
                if (promise && parser.onError) {
                    promise.catch(parser.onError);
                }
            }
            setImmediate(() => {
                if (parser.onChange) {
                    const promise = parser.onChange(parser.ref);
                    if (promise && parser.onError) {
                        promise.catch(parser.onError);
                    }
                }
            });
        }, {
            excludeFiles: config.watchIgnore
        });
        watcher = ts_json_schema_generator_1.ts.createWatchProgram(host);
        try {
            parser.ref = new KitaParser(config, watcher.getProgram().getProgram(), formatter);
        }
        catch (error) {
            const promise = parser.onChange?.(parser.ref);
            if (promise && parser.onError) {
                promise.catch(parser.onError);
            }
        }
        // Make sure to call the onChange hook after the first parse
        setImmediate(() => {
            if (parser.onChange) {
                const promise = parser.onChange(parser.ref);
                if (promise && parser.onError) {
                    promise.catch(parser.onError);
                }
            }
        });
        return parser;
    }
    constructor(config, program, 
    /** A source formatter to call every time a route is generated */
    formatter) {
        this.config = config;
        this.program = program;
        this.formatter = formatter;
        // Json schema
        this.schemaBuilder = new builder_1.SchemaBuilder(this.config, program);
        // Parsing
        this.rootParameterParser = (0, parameter_parsers_1.buildParameterParser)(this.config, this.schemaBuilder, this);
        this.rootProviderParser = (0, provider_parsers_1.buildProviderParser)(this.config, this.rootParameterParser, this.program.getTypeChecker(), this.schemaBuilder);
        this.rootRouteParser = (0, route_parsers_1.buildRouteParser)(this.config, this.schemaBuilder, this.rootParameterParser, this.program.getTypeChecker(), this);
    }
    async *parse() {
        const files = this.program.getSourceFiles().map((f) => f.fileName);
        // Parses all providers first
        for await (const provider of (0, traverser_1.traverseProviders)(this.program, this.rootProviderParser, files.filter((f) => f.startsWith((0, paths_1.toTsPath)(path_1.default.join(this.config.src, common_1.kProvidersFolder)))))) {
            if (provider instanceof common_1.KitaError) {
                yield provider;
                continue;
            }
            if (provider instanceof Error) {
                yield new common_1.UnknownKitaError(provider.message, provider);
                continue;
            }
            // Ignore providers that resolved elsewhere, like in a plugin
            if (!provider) {
                continue;
            }
            const duplicated = this.providers.get(provider.type);
            if (duplicated) {
                yield new common_1.DuplicateProviderTypeError(provider.type, duplicated.providerPath, provider.providerPath);
                continue;
            }
            this.providers.set(provider.type, provider);
        }
        // Parses all routes
        for await (const result of (0, traverser_1.traverseStatements)(this.program, this.rootRouteParser, files.filter((f) => f.startsWith((0, paths_1.toTsPath)(path_1.default.join(this.config.src, common_1.kRoutesFolder)))))) {
            if (result instanceof common_1.KitaError) {
                yield result;
                continue;
            }
            if (result instanceof Error) {
                yield new common_1.UnknownKitaError(result.message, result);
                continue;
            }
            const route = result.parsed;
            const node = result.statement;
            const duplicated = this.routes.get(route.schema.operationId);
            if (duplicated) {
                yield new common_1.DuplicateOperationIdError(route.schema.operationId, 
                //@ts-expect-error - it may not have a name
                node.name || node);
                // Tries to find the duplicated route in the source file
                // but there is no problem if it could not be found
                const duplicatedNode = this.program
                    .getSourceFile(duplicated.relativePath)
                    ?.statements.find((c) => (0, nodes_1.isExportedFunction)(c) && c.name?.getText() === duplicated.controllerMethod);
                if (duplicatedNode) {
                    yield new common_1.DuplicateOperationIdError(duplicated.schema.operationId, 
                    //@ts-expect-error - it may not have a name
                    duplicatedNode.name || duplicatedNode);
                }
                continue;
            }
            this.routes.set(route.schema.operationId, route);
            if (this.formatter) {
                const promise = this.formatter.generateRoute(route, this);
                // Only await if needs to
                if (promise) {
                    await promise;
                }
            }
        }
        if (this.formatter) {
            const promise = this.formatter.generateRuntime(this);
            // Only await if needs to
            if (promise) {
                await promise;
            }
        }
    }
    getProvider(name) {
        return this.providers.get(name);
    }
    getProviders() {
        return Array.from(this.providers.values());
    }
    getProviderCount() {
        return this.providers.size;
    }
    getRoute(operationId) {
        return this.routes.get(operationId);
    }
    getRoutes() {
        return Array.from(this.routes.values());
    }
    getRouteCount() {
        return this.routes.size;
    }
    getSchema(ref) {
        return this.schemaBuilder.getDefinition(ref);
    }
    getSchemas() {
        return this.schemaBuilder.toSchemaArray();
    }
    getSchemaCount() {
        return this.schemaBuilder.getDefinitionCount();
    }
    getPlugin(name) {
        return this.plugins.get(name);
    }
    getPlugins() {
        return Array.from(this.plugins.values());
    }
    getPluginCount() {
        return this.plugins.size;
    }
    addPlugin(name, plugin) {
        this.plugins.set(name, plugin);
    }
    getProgram() {
        return this.program;
    }
}
exports.KitaParser = KitaParser;
//# sourceMappingURL=parser.js.map