﻿"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.DefaultProviderParser = void 0;
const tslib_1 = require("tslib");
const common_1 = require("@kitajs/common");
const path_1 = tslib_1.__importDefault(require("path"));
const ts_json_schema_generator_1 = require("ts-json-schema-generator");
const nodes_1 = require("../util/nodes");
const paths_1 = require("../util/paths");
const traverser_1 = require("../util/traverser");
class DefaultProviderParser {
    paramParser;
    config;
    constructor(paramParser, config) {
        this.paramParser = paramParser;
        this.config = config;
    }
    /** Default provider handles all files. */
    supports() {
        return true;
    }
    async parse(source) {
        const fn = source.statements.find(nodes_1.isDefaultExportFunction);
        if (!fn) {
            throw new common_1.NoProviderExportedError(source);
        }
        if (!fn.type) {
            throw new common_1.UntypedProviderError(fn.name || fn);
        }
        const returnType = (0, nodes_1.unwrapPromiseType)(fn.type);
        const async = fn.type !== returnType;
        const type = (0, nodes_1.getTypeName)(returnType);
        // Not a type reference.
        if (returnType.kind !== ts_json_schema_generator_1.ts.SyntaxKind.TypeReference || !type) {
            throw new common_1.WronglyTypedProviderError(returnType);
        }
        const [schemaTransformer] = this.searchHelper(source, ['transformSchema']);
        // validations
        if (schemaTransformer) {
            this.validateSchemaTransformer(schemaTransformer);
        }
        const applicationHooks = this.searchHelper(source, common_1.ApplicationHookNames);
        for (const hook of applicationHooks) {
            this.validateHooks(hook);
        }
        const lifecycleHooks = this.searchHelper(source, common_1.LifecycleHookNames);
        for (const hook of applicationHooks) {
            this.validateHooks(hook);
        }
        return {
            async,
            type,
            providerPath: (0, paths_1.cwdRelative)(path_1.default.relative(this.config.cwd, source.fileName)),
            schemaTransformer: !!schemaTransformer,
            applicationHooks: applicationHooks.map((h) => h.name.text),
            lifecycleHooks: lifecycleHooks.map((h) => h.name.text),
            parseParameters: async (route, parameterNode) => {
                const parameters = [];
                // Adds all parameters in their respective position
                try {
                    for await (const { param, index } of (0, traverser_1.traverseParameters)(fn, this.paramParser, route)) {
                        parameters[index] = param;
                    }
                }
                catch (error) {
                    if (!(error instanceof common_1.KitaError)) {
                        error = new common_1.UnknownKitaError(String(error), error);
                    }
                    // Only wraps validation errors, other errors should have their stack trace from the
                    // provider file, validation errors should have the trace where they are used.
                    if (error.type === 'validator') {
                        throw new common_1.IncompatibleProviderError(parameterNode.type || parameterNode, error instanceof common_1.RouteParameterMultipleErrors ? error.children : [error]);
                    }
                    throw error;
                }
                return parameters;
            }
        };
    }
    searchHelper(source, fnNames) {
        const hooks = [];
        for (const statement of source.statements) {
            if (!ts_json_schema_generator_1.ts.isFunctionDeclaration(statement)) {
                continue;
            }
            for (const name of fnNames) {
                if ((0, nodes_1.nodeNameEquals)(statement, name)) {
                    hooks.push(statement);
                    break;
                }
            }
        }
        return hooks;
    }
    validateSchemaTransformer(schemaTransformer) {
        if (!(0, nodes_1.isExportedFunction)(schemaTransformer)) {
            throw new common_1.InvalidProviderSchemaTransformerError(schemaTransformer.name || schemaTransformer, 'please export it.');
        }
        if (schemaTransformer.parameters.length < 1) {
            throw new common_1.InvalidProviderSchemaTransformerError(schemaTransformer.name || schemaTransformer, 'it must receive the schema as the first parameter.');
        }
        if (schemaTransformer.parameters.length > 2) {
            throw new common_1.InvalidProviderSchemaTransformerError(schemaTransformer.name || schemaTransformer, 'only RouteSchema and ProviderGenerics are allowed as parameters.');
        }
        if ((0, nodes_1.getTypeNodeName)(schemaTransformer.parameters[0]) !== 'RouteSchema') {
            throw new common_1.InvalidProviderSchemaTransformerError(schemaTransformer.parameters[0].type ||
                schemaTransformer.parameters[0].name ||
                schemaTransformer.name ||
                schemaTransformer, 'the first parameter must be of type `RouteSchema`.');
        }
        // provider generics
        if (schemaTransformer.parameters[1]) {
            if ((0, nodes_1.getTypeNodeName)(schemaTransformer.parameters[1]) !== 'ProviderGenerics') {
                throw new common_1.InvalidProviderSchemaTransformerError(schemaTransformer.parameters[1].type ||
                    schemaTransformer.parameters[1].name ||
                    schemaTransformer.name ||
                    schemaTransformer, 'if the second parameter is present, its type must be `ProviderGenerics`.');
            }
        }
        if (!schemaTransformer.type || (0, nodes_1.getTypeNodeName)(schemaTransformer) !== 'RouteSchema') {
            throw new common_1.InvalidProviderSchemaTransformerError(schemaTransformer.type || schemaTransformer.name || schemaTransformer, 'it must have the `RouteSchema` return type explicitly.');
        }
    }
    validateHooks(hook) {
        if (!(0, nodes_1.isExportedFunction)(hook)) {
            throw new common_1.InvalidProviderHookError(hook.name || hook, 'please export it.');
        }
        if (!hook.modifiers.some((m) => m.kind === ts_json_schema_generator_1.ts.SyntaxKind.AsyncKeyword)) {
            throw new common_1.InvalidProviderHookError(hook.name || hook, 'please it hook async.');
        }
        // TODO: fastify@5.
        // https://github.com/fastify/fastify/issues/4967
        // Parameters are inconsistent across hooks. Until fastify@5 is released,
        // I do not think validating them is worth it.
    }
}
exports.DefaultProviderParser = DefaultProviderParser;
//# sourceMappingURL=default.js.map