﻿"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.HtmlRouteParser = void 0;
const tslib_1 = require("tslib");
const common_1 = require("@kitajs/common");
const path_1 = tslib_1.__importDefault(require("path"));
const ts_json_schema_generator_1 = require("ts-json-schema-generator");
const suspense_id_1 = require("../parameter-parsers/suspense-id");
const jsdoc_1 = require("../util/jsdoc");
const nodes_1 = require("../util/nodes");
const paths_1 = require("../util/paths");
const string_1 = require("../util/string");
const traverser_1 = require("../util/traverser");
class HtmlRouteParser {
    config;
    paramParser;
    checker;
    builder;
    constructor(config, paramParser, checker, builder) {
        this.config = config;
        this.paramParser = paramParser;
        this.checker = checker;
        this.builder = builder;
    }
    supports(node) {
        if (!(0, nodes_1.isExportFunction)(node)) {
            return false;
        }
        const sourceFile = node.getSourceFile();
        if (!node.name || !sourceFile) {
            return false;
        }
        // no .jsx is allowed
        if (!sourceFile.fileName.endsWith('.tsx')) {
            return false;
        }
        // all should not be allowed
        if (!node.name.text.trim().match(/^get|post|put|delete$/i)) {
            return false;
        }
        const defaultExport = node.modifiers.find((s) => s.kind === ts_json_schema_generator_1.ts.SyntaxKind.DefaultKeyword);
        if (defaultExport) {
            throw new common_1.DefaultExportedRoute(defaultExport || node.name || node);
        }
        return true;
    }
    async parse(node) {
        const source = node.getSourceFile();
        const { url, routeId } = (0, string_1.parseUrl)(source.fileName, this.config);
        const method = node.name.getText();
        const route = {
            kind: 'html',
            url,
            controllerMethod: method,
            method: method.toUpperCase(),
            relativePath: (0, paths_1.cwdRelative)(path_1.default.relative(this.config.cwd, source.fileName)),
            parameters: [],
            schema: {
                operationId: `${method.toLowerCase() + routeId}View`,
                hide: true,
                response: { [200]: { type: 'string' } }
            }
        };
        // Parses all jsdoc tags
        (0, jsdoc_1.parseJsDocTags)(node, route);
        const returnType = this.builder.createTypeSchema((0, nodes_1.getReturnType)(node, this.checker));
        const primitive = this.builder.toPrimitive(returnType, true);
        // Non string return type
        if (!primitive || !(primitive instanceof ts_json_schema_generator_1.StringType)) {
            throw new common_1.InvalidHtmlRoute(node.type || node, primitive);
        }
        // Adds all parameters in their respective position
        for await (const { param, index } of (0, traverser_1.traverseParameters)(node, this.paramParser, route)) {
            route.parameters[index] = param;
        }
        const routeParameters = route.parameters.map((r) => r.value).join(', ');
        if (
        // If the SuspenseId parameter was used, we need to render as a stream.
        route.parameters.some((p) => p.name === suspense_id_1.SuspenseIdParameterParser.name)) {
            const boundCall = `${common_1.kControllerName}.${route.controllerMethod}.bind(undefined${routeParameters ? `, ${routeParameters}` : ''})`;
            route.imports ??= [];
            route.imports.push({
                name: '{ renderToStream }',
                path: '@kitajs/html/suspense'
            });
            route.customReturn = `return ${common_1.kReplyParam}.type('text/html; charset=utf-8').send(renderToStream(${boundCall}, ${common_1.kRequestParam}.id));`;
        }
        else {
            const handlerCall = `${common_1.kControllerName}.${route.controllerMethod}.call(undefined${routeParameters ? `, ${routeParameters}` : ''})`;
            route.customReturn = `${common_1.kReplyParam}.type('text/html; charset=utf-8'); return ${handlerCall}`;
        }
        return route;
    }
}
exports.HtmlRouteParser = HtmlRouteParser;
//# sourceMappingURL=html.js.map