﻿"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.RestRouteParser = void 0;
const tslib_1 = require("tslib");
const common_1 = require("@kitajs/common");
const path_1 = tslib_1.__importDefault(require("path"));
const ts_json_schema_generator_1 = require("ts-json-schema-generator");
const helpers_1 = require("../schema/helpers");
const http_1 = require("../util/http");
const jsdoc_1 = require("../util/jsdoc");
const nodes_1 = require("../util/nodes");
const paths_1 = require("../util/paths");
const string_1 = require("../util/string");
const swagger_1 = require("../util/swagger");
const traverser_1 = require("../util/traverser");
class RestRouteParser {
    config;
    schema;
    paramParser;
    typeChecker;
    collector;
    constructor(config, schema, paramParser, typeChecker, collector) {
        this.config = config;
        this.schema = schema;
        this.paramParser = paramParser;
        this.typeChecker = typeChecker;
        this.collector = collector;
    }
    supports(node) {
        if (!(0, nodes_1.isExportFunction)(node)) {
            return false;
        }
        if (!node.name || !node.getSourceFile()) {
            return false;
        }
        // Allows this parameter to not be defined as this is the last
        // resolver and should try to catch as many routes as possible.
        if (!http_1.HttpMethods.includes(node.name.text.toUpperCase()) && node.name.text.toUpperCase() !== 'ALL') {
            return false;
        }
        return true;
    }
    async parse(node) {
        const defaultExport = node.modifiers.find((s) => s.kind === ts_json_schema_generator_1.ts.SyntaxKind.DefaultKeyword);
        if (defaultExport) {
            throw new common_1.DefaultExportedRoute(defaultExport || node.name || node);
        }
        // Adds fastify swagger plugin
        if (!this.collector.getPlugin('fastifySwagger')) {
            this.collector.addPlugin('fastifySwagger', {
                name: 'fastifySwagger',
                importUrl: '@fastify/swagger',
                options: {
                    mode: 'dynamic',
                    openapi: {
                        openapi: '3.1.0',
                        info: swagger_1.DefaultOpenApiInfo
                    },
                    refResolver: {
                        _raw: '{ buildLocalReference(json, _, __, i) { return json.$id || json.$title || json.name || `def-${i}` } }'
                    }
                }
            });
        }
        // Scalar Ui is on a different package
        if (!this.collector.getPlugin('fastifyScalarUi')) {
            this.collector.addPlugin('fastifyScalarUi', {
                name: 'fastifyScalarUi',
                importUrl: '@scalar/fastify-api-reference',
                options: {
                    routePrefix: '/reference',
                    configuration: {
                        theme: 'none',
                        customCss: swagger_1.ScalarUiTheme
                    }
                }
            });
        }
        const source = node.getSourceFile();
        const { url, routeId } = (0, string_1.parseUrl)(source.fileName, this.config);
        const method = node.name.getText();
        const route = {
            kind: 'rest',
            url,
            controllerMethod: method,
            method: method.toUpperCase(),
            relativePath: (0, paths_1.cwdRelative)(path_1.default.relative(this.config.cwd, source.fileName)),
            parameters: [],
            schema: {
                operationId: method.toLowerCase() + routeId,
                // Applies default responses
                response: this.config.responses
            }
        };
        // Parses all jsdoc tags
        (0, jsdoc_1.parseJsDocTags)(node, route);
        // Adds response type.
        try {
            const returnType = this.schema.createTypeSchema((0, nodes_1.getReturnType)(node, this.typeChecker));
            const primitiveReturn = this.schema.toPrimitive(returnType, true);
            if (primitiveReturn instanceof ts_json_schema_generator_1.VoidType || primitiveReturn instanceof ts_json_schema_generator_1.UndefinedType) {
                throw new common_1.RouteWithoutReturnError(node.type || node.name || node);
            }
            (0, helpers_1.mergeSchema)(route, {
                response: {
                    ['2xx']: this.schema.consumeNodeSchema(returnType, (0, common_1.capital)(`${route.schema.operationId}Response`))
                }
            });
        }
        catch (error) {
            if (error instanceof common_1.KitaError) {
                throw error;
            }
            throw new common_1.ReturnTypeError(node.type || node.name || node, error);
        }
        // Adds all parameters in their respective position
        for await (const { param, index } of (0, traverser_1.traverseParameters)(node, this.paramParser, route)) {
            route.parameters[index] = param;
        }
        return route;
    }
}
exports.RestRouteParser = RestRouteParser;
//# sourceMappingURL=rest.js.map