﻿"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.SchemaBuilder = void 0;
const common_1 = require("@kitajs/common");
const ts_json_schema_generator_1 = require("ts-json-schema-generator");
const helpers_1 = require("./helpers");
class SchemaBuilder {
    config;
    /** All definitions read from this schema builder */
    definitions = {};
    parser;
    formatter;
    constructor(config, program) {
        this.config = config;
        const generatorCfg = {
            ...config.generatorConfig,
            tsconfig: config.tsconfig,
            discriminatorType: 'open-api'
        };
        this.parser = (0, ts_json_schema_generator_1.createParser)(program, generatorCfg, (mut) => {
            for (const parser of config.generatorConfig.parsers) {
                mut.addNodeParser(parser);
            }
        });
        this.formatter = (0, helpers_1.removeFormatterDefinitions)((0, helpers_1.correctFormatterChildrenOrder)((0, ts_json_schema_generator_1.createFormatter)(generatorCfg, (mut) => {
            for (const formatter of config.generatorConfig.formatters) {
                mut.addTypeFormatter(formatter);
            }
        })));
    }
    /** Creates a schema for the given ts node */
    createTypeSchema(node) {
        try {
            return this.parser.createType(node, new ts_json_schema_generator_1.Context(node));
        }
        catch (error) {
            throw new common_1.CannotCreateNodeTypeError(node, error);
        }
    }
    /** Saves and returns a {@linkcode ts.Node}'s respective json schema. */
    consumeNodeSchema(node, overrideName) {
        let type = node instanceof ts_json_schema_generator_1.BaseType ? node : this.createTypeSchema(node);
        {
            // Prevents from creating multiple `{ id: '...', type: 'string' }`-like definitions
            const primitive = this.toPrimitive(type);
            if (primitive) {
                return this.formatDefinition(primitive);
            }
        }
        //@ts-expect-error - Defines a return type name to avoid uri-reference problem
        if (overrideName && !type.name)
            type.name ??= overrideName;
        // Includes this node into our recursive definition
        this.appendChildDefinitions(type);
        // @ts-expect-error - unnamed references usually means literal types
        // inside other definitions, so we just unwrap them.
        if (type instanceof ts_json_schema_generator_1.DefinitionType && !type.name) {
            type = type.getType();
        }
        // Returns reference to this node
        return this.formatDefinition(type);
    }
    /**
     * Attempts to unwrap possible type wrappers and resolve into a primitive typescript base type.
     *
     * @param skipWrappers If we should return OptionalType or AnnotatedType wrappers or its inner type
     */
    toPrimitive(type, skipWrappers = false) {
        if (!(type instanceof ts_json_schema_generator_1.BaseType)) {
            type = this.createTypeSchema(type);
        }
        // Primitive types
        if (type instanceof ts_json_schema_generator_1.PrimitiveType ||
            type instanceof ts_json_schema_generator_1.UndefinedType ||
            type instanceof ts_json_schema_generator_1.VoidType ||
            type instanceof ts_json_schema_generator_1.LiteralType) {
            return type;
        }
        // Type wrappers
        if (type instanceof ts_json_schema_generator_1.DefinitionType || type instanceof ts_json_schema_generator_1.ReferenceType || type instanceof ts_json_schema_generator_1.AliasType) {
            return this.toPrimitive(type.getType(), skipWrappers);
        }
        // Wrappers that should return itself
        if (type instanceof ts_json_schema_generator_1.OptionalType || type instanceof ts_json_schema_generator_1.AnnotatedType) {
            const primitive = this.toPrimitive(type.getType(), skipWrappers);
            if (primitive) {
                return skipWrappers ? primitive : type;
            }
            return undefined;
        }
        // Array type can have only one type
        if (type instanceof ts_json_schema_generator_1.ArrayType) {
            if (this.toPrimitive(type.getItem(), skipWrappers)) {
                return type;
            }
            return undefined;
        }
        // Array types that can have multiple types
        if (type instanceof ts_json_schema_generator_1.UnionType ||
            type instanceof ts_json_schema_generator_1.IntersectionType ||
            type instanceof ts_json_schema_generator_1.EnumType ||
            type instanceof ts_json_schema_generator_1.TupleType) {
            const types = type.getTypes();
            // Some types like Promise<T> and T gets resolved into a union of a single type
            if (types.length === 1) {
                return this.toPrimitive(types[0], skipWrappers);
            }
            if (type.getTypes().every((t) => this.toPrimitive(t, skipWrappers))) {
                return type;
            }
        }
        // No primitive type found
        return undefined;
    }
    /** Get definition for a base type without the `#/definitions/` prefix. */
    formatDefinition(type) {
        return this.formatter.getDefinition(type);
    }
    /** Get all definitions as an array. */
    toSchemaArray() {
        return Object.keys(this.definitions).map((name) => this.getDefinition(name));
    }
    /** Get the number of definitions. */
    getDefinitionCount() {
        return Object.keys(this.definitions).length;
    }
    /** Get the definition for a type name */
    getDefinition(name) {
        const definition = this.definitions[name];
        if (!definition) {
            return undefined;
        }
        // Make sure always a definition has a $id property
        if (!definition.$id) {
            // Encode the definition key if the option is enabled
            definition.$id = this.config.generatorConfig.encodeRefs ? encodeURIComponent(name) : name;
        }
        return definition;
    }
    /** Appends new definitions to the schema builder. */
    appendChildDefinitions(type, inner) {
        const seen = new Set();
        const nameIdMap = new Map();
        const definitions = inner ? (this.definitions[inner] ??= {}) : this.definitions;
        for (const child of this.formatter.getChildren(type)) {
            if (!(child instanceof ts_json_schema_generator_1.DefinitionType) || seen.has(child.getId())) {
                continue;
            }
            // No duplicate definitions found
            seen.add(child.getId());
            const name = child.getName();
            const previousId = nameIdMap.get(name);
            // Remove def prefix from ids to avoid false alarms
            const childId = child.getId().replace(/def-/g, '');
            if (previousId && childId !== previousId) {
                throw new common_1.MultipleDefinitionsError(name);
            }
            // No duplicate definitions found
            nameIdMap.set(name, childId);
            // Add child definition
            if (!(name in definitions)) {
                // @ts-expect-error - unnamed child should be kept inside the original definition
                if (!child.name) {
                    continue;
                }
                definitions[name] = this.formatter.getDefinition(child.getType());
            }
        }
    }
}
exports.SchemaBuilder = SchemaBuilder;
//# sourceMappingURL=builder.js.map