﻿"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.unwrapPromiseType = exports.nodeNameEquals = exports.isExportFunction = exports.isDefaultExportFunction = exports.isExportedFunction = exports.isExportedVariable = exports.getReturnType = exports.isParamOptional = exports.hasThisParameter = exports.getParameterGenerics = exports.getParameterName = exports.getIdentifierName = exports.getTypeNodeName = exports.getTypeName = void 0;
const common_1 = require("@kitajs/common");
const ts_json_schema_generator_1 = require("ts-json-schema-generator");
const syntax_1 = require("./syntax");
/** Gets the trimmed type name. */
function getTypeName(type) {
    return type?.getFirstToken()?.getText().trim();
}
exports.getTypeName = getTypeName;
/** Gets the parameter type name. */
function getTypeNodeName(param) {
    return getTypeName(param.type);
}
exports.getTypeNodeName = getTypeNodeName;
/** Gets the parameter type name. */
function getIdentifierName(param) {
    return ts_json_schema_generator_1.ts.isIdentifier(param) ? param.text.trim() : param.getText().trim();
}
exports.getIdentifierName = getIdentifierName;
/**
 * Tries to resolve the parameter name, either by a string literal specified by {@linkcode genericIndex} or the parameter
 * name.
 *
 * @param genericIndex If the provided parameter is a NodeWithTypeArguments, which generic we should use?
 */
function getParameterName(node, genericIndex) {
    const genericNode = getParameterGenerics(node)[genericIndex];
    // The user overridden the name of the parameter with a string literal
    if (genericNode) {
        // @ts-expect-error TODO: Find correct type
        if (genericNode.name && !ts_json_schema_generator_1.ts.isIdentifier(genericNode.name)) {
            throw new common_1.CannotResolveParameterNameError(genericNode);
        }
        return (0, syntax_1.unquote)(genericNode.getText());
    }
    // We may find a parameter with a destructuring pattern or similar syntaxes
    if (!ts_json_schema_generator_1.ts.isIdentifier(node.name)) {
        throw new common_1.CannotResolveParameterNameError(node.name);
    }
    // The user didn't override the name of the parameter, so we use the parameter name
    return getIdentifierName(node.name);
}
exports.getParameterName = getParameterName;
/** Gets the parameter type generics. */
function getParameterGenerics(param) {
    return param.type?.typeArguments || [];
}
exports.getParameterGenerics = getParameterGenerics;
/** If the provided function has a `this` parameter. */
function hasThisParameter(fn) {
    return fn.parameters[0] && getIdentifierName(fn.parameters[0].name) === 'this';
}
exports.hasThisParameter = hasThisParameter;
/** Returns true if the parameter is optional or has a initializer. */
function isParamOptional(param) {
    return !!param.questionToken || !!param.initializer;
}
exports.isParamOptional = isParamOptional;
/** Finds the return type of a declaration or creates one if it doesn't exist. */
function getReturnType(node, typeChecker) {
    if (node.type) {
        return node.type;
    }
    const signature = typeChecker.getSignatureFromDeclaration(node);
    const implicitType = typeChecker.getReturnTypeOfSignature(signature);
    return typeChecker.typeToTypeNode(implicitType, undefined, 
    // There's no documentation on what flags we should use
    // I talked with one of typescript engineers on discord and
    // probably only NoTruncation would be useful. However, more
    // research should be done.
    ts_json_schema_generator_1.ts.NodeBuilderFlags.NoTruncation);
}
exports.getReturnType = getReturnType;
function isExportedVariable(node) {
    return (
    // Is a function type
    ts_json_schema_generator_1.ts.isVariableStatement(node) &&
        // needs to have modifiers
        !!node.modifiers &&
        // `export` modifier needs to be present
        node.modifiers.some((m) => m.kind === ts_json_schema_generator_1.ts.SyntaxKind.ExportKeyword) &&
        // cannot be a default export
        !node.modifiers.some((m) => m.kind === ts_json_schema_generator_1.ts.SyntaxKind.DefaultKeyword));
}
exports.isExportedVariable = isExportedVariable;
/** Returns true if the provided node is a exported function. */
function isExportedFunction(node) {
    return (
    // Is a exported function
    isExportFunction(node) &&
        // cannot be a default export
        !node.modifiers.some((m) => m.kind === ts_json_schema_generator_1.ts.SyntaxKind.DefaultKeyword));
}
exports.isExportedFunction = isExportedFunction;
/** Returns true if the provided node is a default exported function. */
function isDefaultExportFunction(node) {
    return (
    // Is a exported function
    isExportFunction(node) &&
        // has a default export
        node.modifiers.some((m) => m.kind === ts_json_schema_generator_1.ts.SyntaxKind.DefaultKeyword));
}
exports.isDefaultExportFunction = isDefaultExportFunction;
/** Returns true if the provided node is a **default or not exported** function. */
function isExportFunction(node) {
    return (
    // Is a function type
    ts_json_schema_generator_1.ts.isFunctionDeclaration(node) &&
        // needs to have modifiers
        !!node.modifiers &&
        // `export` modifier needs to be present
        node.modifiers.some((m) => m.kind === ts_json_schema_generator_1.ts.SyntaxKind.ExportKeyword));
}
exports.isExportFunction = isExportFunction;
/** Asserts if this node identifier name */
function nodeNameEquals(node, name) {
    return !!node.name && getIdentifierName(node.name) === name;
}
exports.nodeNameEquals = nodeNameEquals;
/** If the provided type is a promise, returns the type of the promise, otherwise returns the type itself. */
function unwrapPromiseType(type) {
    if (ts_json_schema_generator_1.ts.isTypeReferenceNode(type) && type.typeName.getText() === 'Promise') {
        return type.typeArguments?.[0] || type;
    }
    return type;
}
exports.unwrapPromiseType = unwrapPromiseType;
//# sourceMappingURL=nodes.js.map