﻿"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.traverseParameters = exports.traverseProviders = exports.traverseStatements = void 0;
const common_1 = require("@kitajs/common");
/** Traverse each statement of a source file for the given path and yields the result of the provided parser or an error. */
async function* traverseStatements(program, parser, files) {
    for (const file of files) {
        const sourceFile = program.getSourceFile(file);
        if (!sourceFile) {
            yield new common_1.SourceFileNotFoundError(file);
            return;
        }
        let parsed = 0;
        for (const statement of sourceFile.statements) {
            let supports = parser.supports(statement);
            if ((0, common_1.isPromiseLike)(supports)) {
                supports = await supports;
            }
            if (supports) {
                parsed++;
                try {
                    let parsed = parser.parse(statement);
                    if ((0, common_1.isPromiseLike)(parsed)) {
                        parsed = await parsed;
                    }
                    yield { parsed, statement };
                }
                catch (error) {
                    // Also returns errors as values to be handled by the caller
                    // This allows us to keep parsing even if some routes fail
                    if (error instanceof Error) {
                        yield error;
                    }
                    else {
                        yield new common_1.UnknownKitaError(String(error), error);
                    }
                }
            }
        }
        if (parsed === 0) {
            yield new common_1.EmptyRouteFileError(sourceFile);
        }
    }
}
exports.traverseStatements = traverseStatements;
/** Traverse each source file for the given path and yields the result of the provided parser or an error. */
async function* traverseProviders(program, parser, files) {
    files = files.slice();
    const errors = [];
    while (files.length) {
        const initialLength = files.length;
        for (let i = 0; i < initialLength; i++) {
            const filename = files.shift();
            const source = program.getSourceFile(filename);
            if (!source) {
                throw new common_1.SourceFileNotFoundError(filename);
            }
            let supports = parser.supports(source);
            if ((0, common_1.isPromiseLike)(supports)) {
                supports = await supports;
            }
            if (supports) {
                try {
                    let parsed = parser.parse(source);
                    if ((0, common_1.isPromiseLike)(parsed)) {
                        parsed = await parsed;
                    }
                    yield parsed;
                }
                catch (error) {
                    if (
                    // If a parameter resolver is not found, it may be a provider
                    // that would be parsed by the next parser in the chain
                    error instanceof common_1.ParameterResolverNotFoundError) {
                        errors.push(error);
                        files.push(filename);
                        continue;
                    }
                    // Also returns errors as values to be handled by the caller
                    // This allows us to keep parsing even if some routes fail
                    if (error instanceof Error) {
                        yield error;
                    }
                    else {
                        yield new common_1.UnknownKitaError(String(error), error);
                    }
                }
            }
        }
        // All files could be parsed
        if (files.length === 0) {
            break;
        }
        // If the file length is the same, there is no point in trying again
        if (files.length === initialLength) {
            yield* errors;
            break;
        }
        files = files.reverse();
    }
}
exports.traverseProviders = traverseProviders;
/**
 * Traverse each parameter for the given function and yields the result of the provided parser. Errors are thrown
 * synchronously.
 */
async function* traverseParameters(fn, parser, route) {
    let parameterIndex = 0;
    const errors = [];
    for (let index = 0; index < fn.parameters.length; index++) {
        try {
            const parameter = fn.parameters[index];
            let supports = parser.supports(parameter);
            if ((0, common_1.isPromiseLike)(supports)) {
                supports = await supports;
            }
            // All parameters should be supported by at least one parser
            if (!supports) {
                throw new common_1.ParameterResolverNotFoundError(parameter.type || parameter);
            }
            // Yield the result of the parser (promises are unwrapped)
            let param = parser.parse(parameter, route, fn, index);
            if (param instanceof Promise) {
                param = await param;
            }
            // Ignored parameter
            if (param.ignore) {
                continue;
            }
            yield { param, index: parameterIndex++ };
        }
        catch (error) {
            if (error instanceof common_1.KitaError) {
                errors.push(error);
            }
            else {
                errors.push(new common_1.UnknownKitaError(String(error), error));
            }
        }
    }
    if (errors.length > 1) {
        throw new common_1.RouteParameterMultipleErrors(fn.name || fn, errors);
    }
    else if (errors.length) {
        throw errors[0];
    }
}
exports.traverseParameters = traverseParameters;
//# sourceMappingURL=traverser.js.map